"""Convert MS Office documents to PDF."""

import argparse
import logging
import tempfile
import uuid
from pathlib import Path
from typing import List, Optional

import pymupdf
import xmltodict
from spire.doc import Document  # type: ignore
from spire.doc import FileFormat as DocFileFormat
from spire.doc.common import License as DocLicense
from spire.presentation import FileFormat as PptFileFormat  # type: ignore
from spire.presentation import Presentation
from spire.presentation.common import License as PresentationLicense
from spire.xls import Workbook, XlsWorksheet  # type: ignore
from spire.xls.common import License as XlsLicense

log = logging.getLogger(__name__)


SPIRE_LICENSE_PATH = Path("license.elic.xml")
if SPIRE_LICENSE_PATH.exists():
    log.info('Setting Spire Office license file to "%s"', SPIRE_LICENSE_PATH)
    with SPIRE_LICENSE_PATH.open("rb") as file:
        license = xmltodict.parse(file, xml_attribs=True)
    license_key = license["License"]["@Key"]
    DocLicense.SetLicenseKey(license_key)
    PresentationLicense.SetLicenseKey(license_key)
    XlsLicense.SetLicenseKey(license_key)


class OfficeRenderer:
    """Convert MS Office documents to PDF."""

    CSV_EXTENSIONS = [".csv"]
    XLS_EXTENSIONS = [".xls", ".xlsx", ".xlsb", ".xlsm", ".ods"]
    DOC_EXTENSIONS = [".doc", ".docx", ".dot", ".dotx", ".docm", ".ott"]
    PPT_EXTENSIONS = [".ppt", ".pptx"]

    EXTENSIONS = CSV_EXTENSIONS + XLS_EXTENSIONS + DOC_EXTENSIONS + PPT_EXTENSIONS

    def __init__(self, filename: str):
        """."""
        self.filename = filename
        self.temp = tempfile.TemporaryDirectory()
        self.temp_path = Path(self.temp.name)

    def __enter__(self):
        """."""
        return self

    def __exit__(self, type, value, tb):
        """."""
        try:
            self.temp.cleanup()
        except Exception as exc:
            # must not happen
            log.exception(str(exc))

    def create_tempfile(self, suffix: Optional[str] = None) -> str:
        """."""
        return (self.temp_path / f"{str(uuid.uuid4())}{suffix or ''}").as_posix()

    def save_to_pdf(self, pdf_filename: str, print_area: Optional[str] = None) -> None:
        """Save  to PDF.

        Args:
            pdf_filename (str): The PDF filename
            print_area (str: optional): The selection in the Excel workbook. Example: Sheet1!A1:H32
        """
        extension = Path(self.filename).suffix
        if extension in OfficeRenderer.CSV_EXTENSIONS:
            self._render_csv(pdf_filename)
        elif extension in OfficeRenderer.XLS_EXTENSIONS:
            self._render_xls(pdf_filename, print_area)
        elif extension in OfficeRenderer.DOC_EXTENSIONS:
            self._render_doc(pdf_filename)
        elif extension in OfficeRenderer.PPT_EXTENSIONS:
            self._render_ppt(pdf_filename)
        else:
            raise ValueError(f'Unsupported extension: "{extension}')

    def _render_xls(self, pdf_filename: str, print_area: str) -> None:
        """."""
        log.info('Render "%s" to "%s"', self.filename, pdf_filename)
        # Create a Workbook object
        workbook = Workbook()
        doc = pymupdf.open()
        try:
            # Load an Excel document
            workbook.LoadFromFile(self.filename)

            # Set worksheet to fit to page when converting
            # SheetFitToWidth, SheetFitToPage
            workbook.ConverterSetting.SheetFitToWidth = True

            # Select sheet and print area
            sheet: XlsWorksheet
            sheets: List[XlsWorksheet]
            if print_area and "!" in print_area:
                # Summary is a selection in the Excel workbook (print area).
                # Example: NB-maurerkr-5001556-0001-002
                # <summary selection="Sheet1!A1:H32" />
                sheet_name, print_area = print_area.split("!")
                sheet = workbook.Worksheets.get_Item(sheet_name)
                if sheet:
                    sheet.PageSetup.PrintArea = print_area
                sheets = [sheet]
            else:
                # Select all non-empty sheets
                sheets = [s for s in workbook.Worksheets if not s.IsEmpty]

            if not sheets:
                log.warning(
                    'XLS file "%s" is empty. Print area: %s', self.filename, print_area
                )

            # Convert the worksheets to PDF file
            for sheet in sheets:
                sheet.PageSetup.LeftMargin = 0
                sheet.PageSetup.TopMargin = 0
                sheet.PageSetup.RightMargin = 0
                sheet.PageSetup.BottomMargin = 0
                sheet.PageSetup.CenterFooter = ""
                sheet.PageSetup.CenterHeader = ""
                sheet.PageSetup.LeftFooter = ""
                sheet.PageSetup.LeftHeader = ""
                sheet.PageSetup.RightFooter = ""
                sheet.PageSetup.RightHeader = ""
                sheet.PageSetup.IsPrintGridlines = True
                sheet_pdf_file = self.create_tempfile(".pdf")
                sheet.SaveToPdf(sheet_pdf_file)
                doc.insert_file(sheet_pdf_file)

            doc.ez_save(pdf_filename)
        except pymupdf.FileDataError:
            log.exception(pymupdf.TOOLS.mupdf_warnings())
            raise
        except Exception:
            log.exception('Render "%s" to "%s" failed', self.filename, pdf_filename)
            raise
        finally:
            doc.close()
            workbook.Dispose()

    def _render_csv(self, pdf_filename: str, separator: Optional[str] = None) -> None:
        """."""
        log.info('Render "%s" to "%s"', self.filename, pdf_filename)
        # Create a Workbook object
        workbook = Workbook()
        try:
            # Load a Text document
            if not separator:
                with Path(self.filename).open() as file:
                    while True:
                        line = file.readline() or ""
                        separator = "," if "," in line else separator
                        separator = ";" if ";" in line else separator
                        if not line or separator:
                            log.debug("Separator: '%s'", separator)
                            break
                if not separator:
                    msg = f"Separator for file {self.filename} not found."
                    raise ValueError(msg)

            workbook.LoadFromFile(self.filename, separator)

            # Set worksheet to fit to page when converting
            # SheetFitToWidth, SheetFitToPage
            workbook.ConverterSetting.SheetFitToWidth = True

            # Select sheet
            sheet: XlsWorksheet = workbook.Worksheets[0]

            # Convert the worksheet to PDF file
            sheet.PageSetup.LeftMargin = 0
            sheet.PageSetup.TopMargin = 0
            sheet.PageSetup.RightMargin = 0
            sheet.PageSetup.BottomMargin = 0
            sheet.PageSetup.CenterFooter = ""
            sheet.PageSetup.CenterHeader = ""
            sheet.PageSetup.LeftFooter = ""
            sheet.PageSetup.LeftHeader = ""
            sheet.PageSetup.RightFooter = ""
            sheet.PageSetup.RightHeader = ""
            sheet.PageSetup.IsPrintGridlines = True
            sheet.SaveToPdf(pdf_filename)
        except Exception:
            log.exception('Render "%s" to "%s" failed', self.filename, pdf_filename)
            raise
        finally:
            workbook.Dispose()

    def _render_doc(self, pdf_filename: str) -> None:
        """."""
        log.info('Render "%s" to "%s"', self.filename, pdf_filename)
        # Create a Document object
        document = Document()
        try:
            # Load a Word document
            document.LoadFromFile(self.filename)
            document.AcceptChanges()
            document.JPEGQuality = 100

            # Convert the document to PDF file

            # parameters = ToPdfParameterList()
            # parameters.IsEmbeddedAllFonts = True
            # parameters.PdfConformanceLevel = PdfConformanceLevel.Pdf_A2B
            # document.SaveToFile(pdf_filename, parameters)
            document.SaveToFile(pdf_filename, DocFileFormat.PDF)
        except Exception:
            log.exception('Render "%s" to "%s" failed', self.filename, pdf_filename)
            raise
        finally:
            document.Close()

    def _render_ppt(self, pdf_filename: str) -> None:
        """."""
        log.info('Render "%s" to "%s"', self.filename, pdf_filename)
        # Example: NB-salazalb-5000388-0001
        # Create a Presentation object
        presentation = Presentation()
        try:
            # Load a PowerPoint presentation
            presentation.LoadFromFile(self.filename)

            # Convert the presentation to PDF file

            # AttributeError: function 'SaveToPdfOption_set_ContainHiddenSlides' not found
            # presentation.SaveToPdfOption.ContainHiddenSlides = False
            presentation.EmbedTrueTypeFonts = True
            presentation.SaveSubsetFonts = True
            # AttributeError: function 'Presentation_set_SlideCountPerPageForPrint' not found
            # presentation.SlideCountPerPageForPrint = 1
            # presentation.SetPageSize(
            #    float(self.page_setup.page_width),
            #    float(self.page_setup.page_height),
            #    True,
            # )
            presentation.SaveToFile(pdf_filename, PptFileFormat.PDF)
        except Exception:
            log.exception('Render "%s" to "%s" failed', self.filename, pdf_filename)
            raise
        finally:
            presentation.Dispose()


if __name__ == "__main__":
    arg_parser = argparse.ArgumentParser()

    arg_parser.add_argument("input", type=str, help="The input file name or folder")
    arg_parser.add_argument("output", type=str, help="The output file name or folder")

    args = arg_parser.parse_args()

    with OfficeRenderer(args.input) as renderer:
        renderer.save_to_pdf(args.output)
